"use strict"

import view from "./view.js"

class UISurveyMatrixFormViewController extends HTMLElement {

  static get observedAttributes() {
    return ["value"];
  }

  constructor(model) {
    super();
    this.model = model || {};
    this.state = {};
    this.state.connected = false;

    this.model.headers = [];
    this.model.questions = [];
    this.model.instructions = "";
    this.model.name = "";

    //Keeps reference of events with bindings (so we can remove them)
    //see: https://stackoverflow.com/questions/11565471/removing-event-listener-which-was-added-with-bind
    this.event = {};
    this.shadowRoot = this.attachShadow({
      mode: "open"
    });
    this.shadowRoot.appendChild(view.content.cloneNode(true));
    this.view = {};
  }

  //Fires when the element is inserted into the DOM. It's a good place to set
  //the initial role, tabindex, internal state, and install event listeners.
  //
  //NOTE: A user may set a property on an instance of an element, before its
  //prototype has been connected to this class. The _upgradeProperty() method
  //will check for any instance properties and run them through the proper
  //class setters.
  connectedCallback() {
    //Wire views here
    this.view.container = this.shadowRoot.querySelector("#container");
    this.view.nameInput = this.shadowRoot.querySelector("#nameInput");
    this.view.instructionInput = this.shadowRoot.querySelector("#instructionInput");
    this.view.questionInputContainer = this.shadowRoot.querySelector("#questionInputContainer");
    this.view.headerInputContainer = this.shadowRoot.querySelector("#headerInputContainer");
    this.view.requiredSwtich = this.shadowRoot.querySelector("#requiredSwitch");

    //Reference events with bindings
    this.event.onKeyup = this._onKeyup.bind(this);
    this.event.onFocus = this._onFocusBlur.bind(this);
    this.event.onBlur = this._onFocusBlur.bind(this);
    this.event.onUpdate = this._onUpdate.bind(this);

    this.view.instructionInput.addEventListener("keyup", this.event.onKeyup);
    this.view.instructionInput.addEventListener("focus", this.event.onFocus);
    this.view.instructionInput.addEventListener("blur", this.event.onBlur);
    this.view.requiredSwtich.addEventListener("update", this.event.onUpdate);

    this.view.nameInput.addEventListener("keyup", this.event.onKeyup);

    this.questionInputs.forEach(questionInput => {
      questionInput.addEventListener("keyup", this.event.onKeyup);
      questionInput.addEventListener("focus", this.event.onFocus);
      questionInput.addEventListener("blur", this.event.onBlur);
    })

    this.headerInputs.forEach(headerInput => {
      headerInput.addEventListener("keyup", this.event.onKeyup);
      headerInput.addEventListener("focus", this.event.onFocus);
      headerInput.addEventListener("blur", this.event.onBlur);
    })

    // Let the element we're live
    this.state.connected = true;
    this._updateView();
  }

  get headerInputs() {
    const headerInputs = this.shadowRoot.querySelectorAll(".header-input");
    return this._arraiafy(headerInputs);
  }

  get questionInputs() {
    const questionInputs = this.shadowRoot.querySelectorAll(".question-input");
    return this._arraiafy(questionInputs);
  }

  attributeChangedCallback(attrName, oldVal, newVal) {

    switch (attrName) {
      case "value":
        if (newVal !== this.value) {
          this.value = newVal;
        }
        break;
      default:
        throw new Error(`Attribute ${attrName} is not handled, you should probably do that`);
    }
  }

  get value() {
    return this.model;
  }
  set value(value) {
    //Check if attribute matches property value, Sync the property with the
    //attribute if they do not, skip this step if already sync
    if (this.getAttribute("value") !== value) {
      //By setting the attribute, the attributeChangedCallback() function is
      //called, which inturn calls this setter again.
      const stringValue = JSON.stringify(value);
      this.setAttribute("value", stringValue);
      //attributeChangeCallback() implicitly called
      return;
    }

    switch (value.constructor) {
      case String:
        this.model = JSON.parse(value);
        break;
      case Object:
        this.model = value;
        break;
    }

    this._updateView();
    this.dispatchEvent(new CustomEvent("update", {
      detail: this.model
    }));
  }

  get shadowRoot() {
    return this._shadowRoot;
  }
  set shadowRoot(value) {
    this._shadowRoot = value
  }

  _onUpdate(e){
    // Remove attribute
    if(!e.detail){
      delete this.model.required;
    } else{
      this.model.required = true;
    }
    this._updateView();
    this.dispatchEvent(new CustomEvent("update", {detail: this.model}));
  }

  _onKeyup(e) {

    this.model.instructions = this.view.instructionInput.value;
    this.model.name = this.view.nameInput.value;

    this.model.questions = [];
    this.questionInputs.forEach(questionInput => {
      if (questionInput.value === "") return;
      this.model.questions.push(questionInput.value);
    })

    this.model.headers = [];
    this.headerInputs.forEach(headerInput => {
      if (headerInput.value === "") return;
      this.model.headers.push(headerInput.value);
    })

    this._autoExpand(e.target);
    this.dispatchEvent(new CustomEvent("update", {
      detail: this.model
    }));
  }

  _onFocusBlur(e) {
    this._removeUnusedInputs("header");
    this._removeUnusedInputs("question");

    const numHeaders = this.headerInputs.length - 1;
    const numQuestions = this.questionInputs.length - 1;
    const lastHeaderInput = this.headerInputs[numHeaders];
    const lastQuestionInput = this.questionInputs[numQuestions];
    const newHeader = this._newInput("header");
    const newQuestion = this._newInput("question");

    switch (e.target) {

      case lastHeaderInput:
        newHeader.addEventListener("keyup", this.event.onKeyup);
        newHeader.addEventListener("focus", this.event.onFocus);
        newHeader.addEventListener("blur", this.event.onBlur);
        this.view.headerInputContainer.appendChild(newHeader);
        break;

      case lastQuestionInput:
        newQuestion.addEventListener("keyup", this.event.onKeyup);
        newQuestion.addEventListener("focus", this.event.onFocus);
        newQuestion.addEventListener("blur", this.event.onBlur);
        this.view.questionInputContainer.appendChild(newQuestion);
        break;
    }

    this._autoExpand(e.target);
    this.dispatchEvent(new CustomEvent("update", {
      detail: this.model
    }));
  }

  _removeUnusedInputs(type) {
    const items = type === "header" ? this.headerInputs : this.questionInputs;
    const lastIndex = items.length - 1;
    const lastItem = items[lastIndex]
    items.forEach(item => {
      const isEmpty = item.value === "" || item.value === null;
      const isNotLast = item !== lastItem;
      if (item && isEmpty && isNotLast) {
        item.remove();
      }
    })
  }

  _newInput(type) {
    const selector = type === "header" ? ".header-input-template" : ".question-input-template";
    const input = this.shadowRoot.querySelector(selector);
    const inputClone = input.cloneNode();
    inputClone.classList.remove(`${type}-input-template`);
    inputClone.classList.add(`${type}-input`);
    inputClone.hidden = false;
    inputClone.value = "";
    return inputClone;
  }

  _elementify(string) {
    const template = document.createElement("div");
    template.innerHTML = string;
    const element = template.querySelector(":first-child");
    template.remove();
    return element;
  }

  _arraiafy(nodeList) {
    const array = [];
    for (var i = 0; i < nodeList.length; i++) {
      array.push(nodeList[i]);
    }
    return array;
  }

  _updateView(view) {
    //No point in rendering if there isn't a model source, or a view on screen
    if (!this.model || !this.state.connected) {
      return;
    }

    switch (view) {

      default:
        this._removeQuestions()
        this.model.questions.forEach(question => {
          const newQuestion = this._newInput("question");
          newQuestion.addEventListener("keyup", this.event.onKeyup);
          newQuestion.addEventListener("focus", this.event.onFocus);
          newQuestion.addEventListener("blur", this.event.onBlur);
          newQuestion.value = question;
          this.view.questionInputContainer.appendChild(newQuestion);
          newQuestion.focus();
        })

        this._removeHeaders()
        this.model.headers.forEach(header => {
          const newHeader = this._newInput("header");
          newHeader.addEventListener("keyup", this.event.onKeyup);
          newHeader.addEventListener("focus", this.event.onFocus);
          newHeader.addEventListener("blur", this.event.onBlur);
          newHeader.value = header;
          this.view.headerInputContainer.appendChild(newHeader);
          newHeader.focus();
        })

        // Unlike the others, this uses view because its not dynamically populated (others are)
        this.view.instructionInput.value = this.model.instructions;
        this.view.nameInput.value = this.model.name;
        this.view.instructionInput.focus();
        this.view.requiredSwtich.setAttribute("value", this.model.required);
        break;
    }
  }

  autoSizeInputs() {
    this.questionInputs.forEach(question => {
      this._autoExpand(question);
    })
    this.headerInputs.forEach(header => {
      this._autoExpand(header);
    })
    this._autoExpand(this.view.instructionInput);
  }

  _autoExpand(element) {

    if (!element) return;

    // Reset element height
    element.style.height = "inherit";

    // Get the computed styles for the element
    var computed = window.getComputedStyle(element);

    // Calculate the height
    var height = parseInt(computed.getPropertyValue("border-top-width"), 10) +
      parseInt(computed.getPropertyValue("padding-top"), 10) +
      element.scrollHeight +
      parseInt(computed.getPropertyValue("padding-bottom"), 10) +
      parseInt(computed.getPropertyValue("border-bottom-width"), 10);

    element.style.height = height + "px";

  }

  _removeHeaders() {
    const headers = this._arraiafy(this.shadowRoot.querySelectorAll(".header-input"));
    headers.forEach(header => header.remove());
  }

  _removeQuestions() {
    const questions = this._arraiafy(this.shadowRoot.querySelectorAll(".question-input"));
    questions.forEach(question => question.remove());
  }

  _getQuestionView(innerHTML) {
    return this._elementify(`
        <div class="row mt-5 mb-5">
            <div class="col-3 mt-3 mr-3">
                ${innerHTML}
            </div>
        </div>`);
  }

  disconnectedCallback() {
    this._removeEvents()
    this.state.connected = false;
  }

  _removeEvents() {
    this.view.container.removeEventListener("click", this.event.keyup);
  }
}

window.customElements.define("ui-survey-matrix-form", UISurveyMatrixFormViewController);
